import unittest

from saw import DB
from saw import sa


class BaseTestQuery(unittest.TestCase):
    """Base class for query tests classes."""

    def setUp(self):
        """Setup a test database."""
        self.db = DB(name='unittests', engine='sqlite', in_memory=True)
        db = self.db

        class Person(db.Model):
            __tablename__ = 'Person'
            name = sa.Column(u'name',
                            sa.VARCHAR,
                            primary_key=True)

        self.db.create()


class TestInserts(BaseTestQuery):

    def test_insert_with_rows_shortcut(self):

        db = self.db
        Person = db.table.Person
        person = db.rows.query(Person).filter_by(name='Unittest').first()
        assert person is None

        newperson = Person()
        newperson.name = 'Unittest'
        db.rows.insert(newperson)
        person = db.rows.query(Person).filter_by(name='Unittest').first()
        assert person.name == 'Unittest'

    def test_insert_with_autosession_context_manager(self):

        db = self.db
        Person = db.table.Person
        person = db.rows.query(Person).filter_by(name='Unittest').first()
        assert person is None

        newperson = Person()
        newperson.name = 'Unittest'

        with db.autosession() as session:
            session.add(newperson)
            session.commit()

        person = db.rows.query(Person).filter_by(name='Unittest').first()
        assert person.name == 'Unittest'

    def test_classic_insert(self):
        """Insert a registry opening a session, adding the new object to the
        session and closing the session."""

        db = self.db
        Person = db.table.Person
        person = db.rows.query(Person).filter_by(name='Unittest').first()
        assert person is None

        newperson = Person()
        newperson.name = 'Unittest'

        session = db.session()
        session.add(newperson)
        session.commit()
        session.close()

        person = db.rows.query(Person).filter_by(name='Unittest').first()
        assert person.name == 'Unittest'



if __name__ == '__main__':
    unittest.main()
