import unittest

from saw import DB
from saw import sa


class BaseTestQuery(unittest.TestCase):
    """Base class for query tests classes."""

    def setUp(self):
        """Setup a test database."""
        self.db = DB(name='unittests', engine='sqlite', in_memory=True)
        db = self.db

        class Person(db.Model):
            __tablename__ = 'Person'
            name = sa.Column(u'name',
                            sa.VARCHAR,
                            primary_key=True)

        self.db.create()


class TestQuery(BaseTestQuery):
    """Test when we make queries."""

    def test_context_manager_query(self):
        """Lets query the db using db.autosession context manager."""

        db = self.db
        Person = db.table.Person

        with db.autosession() as session:
            persons = session.query(Person).all()

        self.assertEquals(len(persons), 0)

    def test_rows_shortcut_query(self):
        """Lets query the db using the rows shortcut."""

        db = self.db
        Person = db.table.Person
        persons = db.rows.query(Person).all()
        self.assertEquals(len(persons), 0)

    def test_classic_query(self):
        """Query opening a session, making the query and then closing
        the session."""

        db = self.db
        Person = db.table.Person

        session = db.session()
        persons = session.query(Person).all()
        session.close()
        self.assertEquals(len(persons), 0)


class TestQueryWithFilters(BaseTestQuery):
    """Test queries with filters."""

    def test_filter_by_name(self):
        """Add a person, then query for it."""

        db = self.db
        Person = db.table.Person

        person = Person()
        person.name = 'Unittest'
        db.rows.insert(person)

        persons = db.rows.query(Person).filter_by(name='Foo').all()
        self.assertEquals(len(persons), 0)

        persons = db.rows.query(Person).filter_by(name='Unittest').all()
        self.assertEquals(len(persons), 1)


if __name__ == '__main__':
    unittest.main()
